//----------------------------------------------
// Night Bird MAIN process
//----------------------------------------------
#include <graphics.h>   
#include <stdio.h>
#include <io.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <conio.h>
#include <ctype.h>
#include <dos.h>
#include <dir.h>
#include <bios.h>
#include <opus.h>
#include <op_graph.h>
#include <op_view.h>
#include <op_color.h>
#include <op_timer.h>
#include "nb_app.h"

Graphics *graphics;      // Declare a graphics object
Viewer *viewer;          // Declare a 3-D viewer object
#ifdef BIRD
Bird *bird;              // Declare a bird object
#endif
#ifdef GROUND
Ground *ground;          // Declare a ground detail object
#endif
Environ *env;            // System environment record
#ifdef TREES
Forest *forest;          // Forest object
#endif
BIOS_Timer *timer;       // Timer object

// System globals
int change_freq;
int current_view;
int view_cnt;
int orbit_dur;
int orbit_half;
int orbit_cnt;
int first_orbit;
int exiting;
int frame_rate;
long letter_z_limit;
long bird_x, bird_y, bird_z;
int tree_cnt;
long frame_check;
long frame_goal;
double actual_rate;
int bird_speed;
int copyright;
int message_display;

#ifdef BENCHMARK
int bench_cnt = 0;
#define BENCH_GOAL 1000
#endif
#ifdef KEY_COMMANDS
#define MIN_BIRD_SPEED   30
#define MAX_BIRD_SPEED  200
#define MIN_VIEW_ELEV    40
#define MAX_VIEW_ELEV  3000
#define MIN_VIEW_DIST   240
#define MAX_VIEW_DIST  1500
#define MIN_VIEW_ZOOM    20
#define MAX_VIEW_ZOOM   200
#define MIN_BIRD_ELEV    40
#define MAX_BIRD_ELEV  2500
int message_frames;
int message_x;
char limit_message[60];
double last_message_occurence;
#define MIN 0
#define MAX 1
#endif

// Local functions
void animate(void);
void opening_message(void);
void prepare(void);
void roll_animation(void);

//-------------------------------------------------------------------------
// Program main
//-------------------------------------------------------------------------
void main(int argc, char **argv)
{
  // Check system
  require_vga_386();

  // Default options
  frame_rate = FRAME_RATE;
  frame_check = 0;
  copyright = ON;

  // Read the command line
  if (argc > 1)
  {
    for (int i = 1; i <= argc; i++)
    {
      strupr(argv[i]);
      if (!strncmp(argv[i], "-F", 2) && strlen(argv[i]) > 2)
        frame_rate = atoi(&argv[i][2]);
      if (!strcmp(argv[i], "-C"))
        copyright = OFF;
    }
  }

  // Initialize
  prepare();

  // Run the animation
  animate();

  // Return to text mode
  graphics->off();

  // Free memory
  #ifdef TREES
  delete forest;
  #endif
  delete env;
  #ifdef GROUND
  delete ground;
  #endif
  #ifdef BIRD
  delete bird;
  #endif
  delete viewer;
  delete graphics;
  delay(300);
}

//-------------------------------------------------------------------------
// Initialize all program settings and objects
//-------------------------------------------------------------------------
void prepare()
{
  // Program scaling factors
  exiting = OFF;

  #ifdef BENCHMARK
  srand(1);
  #else
  randomize();
  #endif

  // View control
  change_freq = CHANGE_MIN + CHANGE_RANGE;
  current_view = HOVER;
  view_cnt = 0;
  bird_speed = BIRD_SPEED;
  frame_goal = frame_rate * 2;
  first_orbit = ON;

  // Create the GRAPHICS and VIEWER objects
  if (!(graphics = new Graphics))
    bail("Mem:graphics");
  if (!(viewer = new Viewer))
    bail("Mem:viewer");

  // Start graphics
  registerbgidriver(EGAVGA_driver);
  registerbgifont(small_font);
  graphics->on(VGA, VGAMED, "");
  delay(500);  // Give auto-sync screens a chance to settle down

  // Set the color palette
  setallpalette(&earth);

  // Define a viewing window - there MUST be at least one
  graphics->define_port(FULL_SCREEN, 0, 0, getmaxx(), getmaxy(), PAGING_ON);
  graphics->define_port(VIEWPORT, 0, 0, getmaxx(), getmaxy(), PAGING_ON);
  graphics->paging_on(SEG_ADR, BGI_PAGING);  // Enable animation
  graphics->reset_pages();
  graphics->set_port(VIEWPORT);

  // Display ownership info
  if (copyright)
  {
    opening_message();
    graphics->chg_active_page();
    graphics->clear_port(VIEWPORT, BLACK);
    graphics->chg_active_page();
    graphics->clear_port(VIEWPORT, BLACK);
  }
  graphics->chg_active_page();
  graphics->clear_port(VIEWPORT, BLACK);
  graphics->chg_active_page();
  graphics->clear_port(VIEWPORT, BLACK);

  // The 3-D zoom setting MUST take place AFTER
  // the above calls to define_port and set_port
  viewer->set_zoom((int)PROG_ZOOM);

  // Create the required objects
  bird_x = bird_z = 0L;
  bird_y = BIRD_ELEV;  
  #ifdef BIRD
  if ((bird = new Bird(bird_x, bird_y, bird_z)) == NULL)
    bail("Mem:bird");
  #endif
  #ifdef GROUND
  if ((ground = new Ground) == NULL)
    bail("Mem:ground");
  #endif
  if (!(env = new Environ(viewer, graphics, 0L, TOT_PAGES, TOT_WINDOWS)))
    bail("Mem:env");
  #ifdef TREES
  if (!(forest = new Forest(env, MAX_TREES, TREE_LINE, Z_HORIZON)))
    bail("Mem:forest");
  #endif

  #ifdef TREES
  #ifdef BENCHMARK
  // Add all of the trees if BENCHMARK
  for (int t = 0; t < forest->tree_tot; t++)
    forest->add_tree();
  #else
  // Start with one tree
  forest->add_tree();
  #endif
  #endif

  // Read and scale the bird settings
  #ifdef BIRD
  bird->scale_settings();
  #endif

  // Set the fill style for erasing with the bounding box
  setfillstyle(SOLID_FILL, EARTH_BLACK);
}

#ifdef PRINT_STATS
//-------------------------------------------------------------------------
// Display Information
//-------------------------------------------------------------------------
void print_info(int state)
{
  int lh = 12;
  int x = 10;
  int y = 10;
  int color = state ? EARTH_MED_CYAN : EARTH_BLACK;

  settextstyle(SMALL_FONT, HORIZ_DIR, 4);
  graphics->print(x, y += lh, color, "Points: %3d of %3d", 
                  ground->get_count(), ground->max_points);
  #ifdef TREES
  graphics->print(x, y += lh, color, "Trees: %2d of %2d",
                  forest->active_trees, forest->tree_tot);
  #endif
  #ifdef BIRD
  graphics->print(x, y += lh, color, "Bird_y: %5ld", bird->center.y);
  #endif
  graphics->print(x, y += lh, color, "Zoom: %3d", (int)viewer->get_zoom());
  x += 560;
  y = 10;
  graphics->print(x, y += lh, color, "X: %6ld", viewer->get_x());
  graphics->print(x, y += lh, color, "Y: %6ld", viewer->get_y());
  graphics->print(x, y += lh, color, "Z: %6ld", viewer->get_z());
  graphics->print(x, y += lh, color, "F: %6d", (int)actual_rate);
}
#endif

//-------------------------------------------------------------------------
// Notify user that a setting has reached its limit
//-------------------------------------------------------------------------
void show_limit(int which, char *setting)
{
  if (!message_frames)
  {
    int color = which == MIN ? EARTH_LT_GREEN : EARTH_WHITE;

    // Build the message
    if (which == MIN)
     sprintf(limit_message, "MINIMUM %s.", setting);
    else
     sprintf(limit_message, "MAXIMUM %s.", setting);

    // Center the message
    settextstyle(SMALL_FONT, HORIZ_DIR, 4);
    message_x = 320 - (textwidth(limit_message) / 2);
    if (message_x & 1)
     message_x--;
   
    // Display the message
    graphics->chg_active_page();
    graphics->print(message_x, 10, color, limit_message);
    graphics->chg_active_page();
    graphics->print(message_x, 10, color, limit_message);
 
    // Sound a warning
    sound(200); delay(20); nosound();
 
    // Start the message frame counter rolling
    message_frames = 20;
 
    // Absorb any further keystrokes
	 while (kbhit()) bioskey(0);
  }
}

//-------------------------------------------------------------------------
// Run the program animation
//-------------------------------------------------------------------------
void animate()
{

  // Place the viewer at a position in space: x, y, z
  viewer->set_position(ST_X, CRUISING_ELEV, ST_Z);  

  // Set the center of interest to be the middle of the bird
  viewer->set_center(bird_x, bird_y, bird_z + 30);
  viewer->view_center();  // Look at the center of interest

  // Create the initial 3-d calculation matrix
  viewer->new_view(); 

  // Draw the scene on both pages
  graphics->chg_active_page();
  #ifdef GROUND
  ground->draw();
  #endif
  #ifdef BIRD
  bird->draw();
  #endif
  #ifdef TREES
  forest->translate();
  forest->draw();
  #endif
  graphics->chg_active_page();
  #ifdef GROUND
  ground->draw();
  #endif
  #ifdef BIRD
  bird->draw();
  #endif
  #ifdef TREES
  forest->translate();
  forest->draw();
  #endif

  // Read the current time
  timer = new BIOS_Timer;
  timer->reset();
  timer->start();

  // Roll the animation
  while (!exiting)
    roll_animation();

  delete timer;
}

//------------------------------------------------------------------------
// This is the actual animation loop
//------------------------------------------------------------------------
void roll_animation()
{
  // Erase the previous scene
  graphics->chg_active_page();
  #ifdef BIRD
  bird->erase();
  #endif
  #ifdef GROUND
  ground->erase();
  #endif
  #ifdef TREES
  forest->erase();
  #endif
  #ifdef STORY
  story->erase();
  #endif
  #ifdef USER_MESSAGE
  user_message->erase();
  #endif

  #ifdef KEY_COMMANDS
  // Process messages
  if (message_frames)
  {
    if (--message_frames == 0)
    {
      graphics->chg_active_page();
      graphics->print(message_x, 10, BLACK, limit_message);
      graphics->chg_active_page();
      graphics->print(message_x, 10, BLACK, limit_message);
		while (kbhit()) bioskey(0);
    }
  }
  else
  #endif

  // Process keyboard events
  if (kbhit())
  {
	 // Process commands
	 double curr_dist;
	 char ch = bioskey(1);
	 while (kbhit()) bioskey(0);
	 if (ch == 27)
      exiting = ON;

    #ifdef KEY_COMMANDS
    int view_changed = OFF;
    switch (toupper(ch))
    {
      #ifdef BIRD
      case 'L':
        if (bird->center.y > MIN_BIRD_ELEV)
        {
          bird->elevation(DOWN, 10);
          viewer->set_center(viewer->get_cx(), 
                             bird->center.y, viewer->get_cz());
          viewer->view_center();
          view_changed = ON;
        }
        else
          show_limit(MIN, "bird height");
        break;
      case 'H':
        if (bird->center.y < MAX_BIRD_ELEV)
        {
          bird->elevation(UP, 10);
          viewer->set_center(viewer->get_cx(), 
                             bird->center.y, viewer->get_cz());
          viewer->view_center();
          view_changed = ON;
        }
        else
          show_limit(MAX, "bird height");
        break;
      case 'B':
        if (bird_speed > MIN_BIRD_SPEED)
          bird_speed -= 5;
        else
          show_limit(MIN, "bird speed");
        break;
      case 'S':
        if (bird_speed < MAX_BIRD_SPEED)
          bird_speed += 5;
        else
          show_limit(MAX, "bird speed");
        break;
      #endif
      case 'D':
        if (viewer->get_y() > MIN_VIEW_ELEV)
        {
          viewer->move_down(10);
          view_changed = ON;
        }
        else
          show_limit(MIN, "viewer height");
        break;
      case 'U':
        if (viewer->get_y() < MAX_VIEW_ELEV)
        {
          viewer->move_up(10);
          view_changed = ON;
        }
        else
          show_limit(MAX, "viewer height");
        break;
      case 'C':
        if (viewer->get_distance() > MIN_VIEW_DIST)
        {
          viewer->move_forward(10);
          view_changed = ON;
        }
        else
          show_limit(MIN, "viewer distance.");
        break;
      case 'F':
        curr_dist = viewer->calc_distance();
        if (curr_dist < MAX_VIEW_DIST)
        {
          viewer->move_forward(-10);
          view_changed = ON;
        }
        else
          show_limit(MAX, "viewer distance");
        break;
      case 'W':
        if (viewer->get_zoom() > MIN_VIEW_ZOOM)
        {
          viewer->zoom_out(1);
          view_changed = ON;
        }
        else
          show_limit(MIN, "viewing magnification");
        break;
      case 'Z':
        if (viewer->get_zoom() < MAX_VIEW_ZOOM)
        {
          viewer->zoom_in(1);
          view_changed = ON;
        }
        else
          show_limit(MAX, "viewing magnification");
        break;
       #ifdef PRINT_STATS
       case 'I':
          graphics->chg_active_page();
          print_info(ON);
          graphics->chg_active_page();
          print_info(ON);
          break;
       case 'O':
          graphics->chg_active_page();
          print_info(OFF);
          graphics->chg_active_page();
          print_info(OFF);
          break;
       #endif
    }
    if (view_changed)
    {
      viewer->view_center();
      viewer->new_view();
    }
    #endif // KEY_COMMANDS
  }

  // Make changes to the scene
  #ifdef BIRD
  bird->flap_wings();
  #endif
  #ifdef GROUND
  ground->move(bird_speed);
  #endif
  #ifdef TREES
  forest->move(bird_speed);
  #endif

  #ifdef VIEW_CHANGER
  // Process changes to the viewer position
  if (current_view == ORBIT)
  {
    int chg;
    double increment = (double)ORBIT_SPEED / orbit_half;

    if (orbit_cnt < orbit_half)
      chg = increment * orbit_cnt;
    else
      chg = increment * (orbit_dur - orbit_cnt);

    viewer->orbit(chg);
    viewer->chg_heading(chg);
    viewer->new_view();
    if (orbit_cnt++ == orbit_dur)
    {
      view_cnt = 0;
      current_view = HOVER;
      if (first_orbit)
        first_orbit = OFF;
    }
  }
  #endif // VIEW_CHANGER

  // Redisplay the scene
  #ifdef GROUND
  ground->draw();
  #endif

  #ifdef TREES
  forest->translate();
  forest->draw_back();
  #ifdef BIRD
  bird->draw();
  #endif
  forest->draw_front();
  #else
  #ifdef BIRD
  bird->draw();
  #endif
  #endif

  graphics->reset();
  graphics->chg_visual_page();

  // Update the view changer
  #ifdef VIEW_CHANGER
  if (++view_cnt == change_freq)
  {
    current_view = ORBIT;
    view_cnt = 0;
    change_freq = CHANGE_MIN + random(CHANGE_RANGE);

    if (current_view == ORBIT)
    {
      orbit_cnt = 0;
      if (first_orbit)
        orbit_dur = 32;
      else
        orbit_dur = ORBIT_MIN + random(ORBIT_RANGE);
      orbit_half = orbit_dur / 2;
    }
  }
  #endif

  // Maintain the timer if still running
  #ifndef BENCHMARK

  // Periodically check the frame rate
  if (++frame_check == frame_goal)
  {
    timer->stop();
    actual_rate = ((double)frame_check / timer->time());
    frame_check = 0;
    timer->reset();

    // Add more detail if the frame rate is too fast
    if (actual_rate > frame_rate)
    {
      #ifdef GROUND
      if (ground->get_count() < ground->max_points)
        ground->add_points();
      #endif

      #ifdef TREES
      forest->add_tree();
      #endif
    }
  }

  #else // BENCHMARK

  if (++bench_cnt == BENCH_GOAL)
  {
    timer->stop();
    graphics->print(10, 10, WHITE, "%5.5f", timer->time());
	 bioskey(0);
    exiting = TRUE;
  }
  #endif

  // Add a delay to reduce flicker
  delay(8);
}
