//----------------------------------------------
// BIRD object
//----------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <graphics.h>
#include <opus.h>
#include <op_graph.h>
#include <op_view.h>
#include <op_color.h>
#include "nb_app.h"

#ifdef BIRD
//
// Bird dimensions in INCHES
//
#define SHOULDER_WIDTH   2.6
#define TAIL_WIDTH       4.7
#define WING_SPAN_1      2.7
#define WING_SPAN_2      3.5
#define WING_SPAN_3      8.0
#define WING_WIDTH_F1    0.1
#define WING_WIDTH_F2    0.3
#define TAIL_START       4.8
#define TAIL_EXTRA       0.5
#define WING_WIDTH_B1    3.6
#define WING_WIDTH_B2    3.5
#define WING_WIDTH_B3    3.4
#define WING_WIDTH_B4    4.1
#define HEAD_LENGTH      2.4
#define BODY_LENGTH      7.5
#define HEAD_HEIGHT      0.5
#define BEAK_HEIGHT     -0.25
#define THROAT_HEIGHT   -1.0
#define TAIL_HEIGHT      0.1
#define BODY_SHIFT       1.2

#define NEAR_ST_ANGLE    240
#define NEAR_SPEED       1.7
#define NEAR_SPREAD      1.7 
#define MID_ST_ANGLE     240
#define MID_SPEED        1.7
#define MID_SPREAD       1.8
#define TIP_ST_ANGLE      50
#define TIP_SPEED        1.9
#define TIP_SPREAD       2.0

//-------------------------------------------------------------------------
// Bird class constructor
//-------------------------------------------------------------------------
Bird::Bird(long hx, long hy, long hz)
{
  // Determine a random wing movement
  flap_direction = UP;
  flap_cnt = random(FLAP_STEPS);

  // Define the BIRD's center point
  set_center(hx, hy, hz);

  // Initialize a bounding box
  st_box[0].x = getmaxx();
  st_box[0].y = getmaxy();
  st_box[1].x = st_box[1].y = 0;
}

//-------------------------------------------------------------------------
// Accepts bird dimensions from the designer and multiplies them
// by the scene scaling factor to equal TENTHS of an INCH
//-------------------------------------------------------------------------
void Bird::scale_settings()
{
  double factor = 10 * BIRD_SIZE;  // INCHES to TENTH_OF_INCHES

  shoulder_width = (long)(factor * SHOULDER_WIDTH);
  tail_width = (long)(factor * TAIL_WIDTH);
  wing_span_1 = (long)(factor * WING_SPAN_1);
  wing_span_2 = (long)(factor * WING_SPAN_2);
  wing_span_3 = (long)(factor * WING_SPAN_3);
  wing_width_f1 = (long)(factor * WING_WIDTH_F1);
  wing_width_f2 = (long)(factor * WING_WIDTH_F2);
  wing_width_b1 = (long)(factor * WING_WIDTH_B1);
  wing_width_b2 = (long)(factor * WING_WIDTH_B2);
  wing_width_b3 = (long)(factor * WING_WIDTH_B3);
  wing_width_b4 = (long)(factor * WING_WIDTH_B4);
  head_length = (long)(factor * HEAD_LENGTH);
  body_length = (long)(factor * BODY_LENGTH);
  tail_start = (long)(factor * TAIL_START);
  tail_extra = (long)(factor * TAIL_EXTRA);
  head_height = (long)(factor * HEAD_HEIGHT);
  beak_height = (long)(factor * BEAK_HEIGHT);
  throat_height = (long)(factor * THROAT_HEIGHT);
  tail_height = (long)(factor * TAIL_HEIGHT);

  // Set the body shift increment
  shift_inc = (long)(factor * BODY_SHIFT) / FLAP_STEPS;

  // Assign the mid-point of the wing flapping
  flap_mid = (FLAP_STEPS - 1) / 2;

  // Calculate the angle increment of the wing segments
  near_flap_inc = ((900 - NEAR_ST_ANGLE) * 2) / (FLAP_STEPS - 1);
  mid_flap_inc = ((900 - MID_ST_ANGLE) * 2) / (FLAP_STEPS - 1);
  tip_flap_inc = ((900 - TIP_ST_ANGLE) * 2) / (FLAP_STEPS - 1);

  // Define the bird using the supplied dimensions
  define();
}

//-------------------------------------------------------------------------
// Calculates all bird vertices relative to center
//-------------------------------------------------------------------------
void Bird::define()
{
  int i, angle;
  int d_inc, st_angle;
  long cx = center.x;
  long cy = center.y;
  long cz = center.z;

  // Brain
  vert[0].world.x = cx;
  vert[0].world.y = cy + head_height;
  vert[0].world.z = cz;

  // Throat point
  vert[1].world.x = cx;
  vert[1].world.y = cy + throat_height;
  vert[1].world.z = cz;

  // Beak point
  vert[2].world.x = cx;
  vert[2].world.y = cy + beak_height;
  vert[2].world.z = cz - head_length;

  // Right Shoulder
  vert[3].world.x = cx + (shoulder_width / 2);
  vert[3].world.y = cy;
  vert[3].world.z = cz;

  // Left Shoulder
  vert[15].world.x = cx - (shoulder_width / 2);
  vert[15].world.y = cy;
  vert[15].world.z = cz;

  // Body end-point
  vert[5].world.x = cx;
  vert[5].world.y = cy;
  vert[5].world.z = cz + body_length;

  // Tail start
  vert[6].world.x = cx;
  vert[6].world.y = cy;
  vert[6].world.z = cz + tail_start;

  // Right tail tip
  vert[7].world.x = cx + (tail_width / 2);
  vert[7].world.y = cy + tail_height;
  vert[7].world.z = cz + body_length + tail_extra;

  // Left tail tip
  vert[8].world.x = cx - (tail_width / 2);
  vert[8].world.y = cy + tail_height;
  vert[8].world.z = cz + body_length + tail_extra;

  // Wing rear at body Right
  double side_width = vert[3].world.x - vert[5].world.x;
  double side_length = vert[5].world.z - vert[3].world.z;
  double side_ratio = side_width / side_length;
  long mid_point = vert[3].world.x - (side_ratio * wing_width_b1);
  vert[4].world.x = cx + mid_point;
  vert[4].world.y = cy;
  vert[4].world.z = cz + wing_width_b1;

  // Wing rear at body Left
  vert[14].world.x = cx - mid_point;
  vert[14].world.y = cy;
  vert[14].world.z = cz + wing_width_b1;

  //
  // The wing vertices only need to be initialized
  // with their Z components because their X and Y
  // components are taken from the wing positions
  // table, and assigned in the function set_wings()
  //
  vert[9].world.z = cz + wing_width_f1;    // Wing front near body Right
  vert[16].world.z = cz + wing_width_f1;   // Wing front near body Left
  vert[11].world.z = cz + wing_width_f2;   // Wing front near tip Right
  vert[18].world.z = cz + wing_width_f2;   // Wing front near tip Left
  vert[10].world.z = cz + wing_width_b2;   // Wing rear near body Right
  vert[17].world.z = cz + wing_width_b2;   // Wing rear near body Left
  vert[12].world.z = cz + wing_width_b3;   // Wing rear near tip Right
  vert[19].world.z = cz + wing_width_b3;   // Wing rear near tip Left
  vert[13].world.z = cz + wing_width_b4;   // Wing tip Right
  vert[20].world.z = cz + wing_width_b4;   // Wing tip Left

  // Define the object polygons as a list of
  // vertex indices taken counter-clockwise
  poly[0].v_index[0] = 2;
  poly[0].v_index[1] = 0;
  poly[0].v_index[2] = 3;

  poly[1].v_index[0] = 2;
  poly[1].v_index[1] = 15;
  poly[1].v_index[2] = 0;

  poly[2].v_index[0] = 1;
  poly[2].v_index[1] = 2;
  poly[2].v_index[2] = 3;

  poly[3].v_index[0] = 1;
  poly[3].v_index[1] = 15;
  poly[3].v_index[2] = 2;

  poly[4].v_index[0] = 0;
  poly[4].v_index[1] = 5;
  poly[4].v_index[2] = 3;

  poly[5].v_index[0] = 0;
  poly[5].v_index[1] = 15;
  poly[5].v_index[2] = 5;

  poly[6].v_index[0] = 1;
  poly[6].v_index[1] = 3;
  poly[6].v_index[2] = 5;

  poly[7].v_index[0] = 1;
  poly[7].v_index[1] = 5;
  poly[7].v_index[2] = 15;

  poly[8].v_index[0] = 6;
  poly[8].v_index[1] = 7;
  poly[8].v_index[2] = 8;

  poly[9].v_index[0] = 3;
  poly[9].v_index[1] = 4;
  poly[9].v_index[2] = 10;

  poly[10].v_index[0] = 3;
  poly[10].v_index[1] = 10;
  poly[10].v_index[2] = 9;

  poly[11].v_index[0] = 15;
  poly[11].v_index[1] = 16;
  poly[11].v_index[2] = 17;

  poly[12].v_index[0] = 15;
  poly[12].v_index[1] = 17;
  poly[12].v_index[2] = 14;

  poly[13].v_index[0] = 9;
  poly[13].v_index[1] = 10;
  poly[13].v_index[2] = 12;

  poly[14].v_index[0] = 9;
  poly[14].v_index[1] = 12;
  poly[14].v_index[2] = 11;

  poly[15].v_index[0] = 16;
  poly[15].v_index[1] = 18;
  poly[15].v_index[2] = 19;

  poly[16].v_index[0] = 16;
  poly[16].v_index[1] = 19;
  poly[16].v_index[2] = 17;

  poly[17].v_index[0] = 11;
  poly[17].v_index[1] = 12;
  poly[17].v_index[2] = 13;

  poly[18].v_index[0] = 18;
  poly[18].v_index[1] = 20;
  poly[18].v_index[2] = 19;

  // Calculate the NEAR wing positions
  st_angle = NEAR_ST_ANGLE;
  d_inc = near_flap_inc;
  for (i = 0, angle = st_angle; i < FLAP_STEPS; i++)
  {
    fixed cos_angle, sin_angle;
    viewer->cos_sin(angle, &cos_angle, &sin_angle);
    fixed radius = I2F(wing_span_1 - vert[3].world.x);

    pos[RIGHT][NEAR][i].x = vert[3].world.x +
                            F2I(f_mul(radius, sin_angle));
    pos[RIGHT][NEAR][i].y = cy + F2I(f_mul(radius, cos_angle));
    pos[LEFT][NEAR][i].x = vert[15].world.x -
                           F2I(f_mul(radius, sin_angle));
    pos[LEFT][NEAR][i].y = cy + F2I(f_mul(radius, cos_angle));

    double factor = (double)((FLAP_STEPS) - (abs(i - flap_mid) * NEAR_SPEED)) / 10;
    angle += (d_inc * factor) * NEAR_SPREAD;
  }

  // Calculate the MID wing positions
  st_angle = MID_ST_ANGLE;
  d_inc = mid_flap_inc;
  for (i = 0, angle = st_angle; i < FLAP_STEPS; i++)
  {
    fixed cos_angle, sin_angle;
    viewer->cos_sin(angle, &cos_angle, &sin_angle);

    fixed radius = I2F(wing_span_2 - wing_span_1);
    pos[RIGHT][MID][i].x = pos[RIGHT][NEAR][i].x +
                           F2I(f_mul(radius, sin_angle));
    pos[RIGHT][MID][i].y = pos[RIGHT][NEAR][i].y +
                           F2I(f_mul(radius, cos_angle));
    pos[LEFT][MID][i].x = pos[LEFT][NEAR][i].x -
                          F2I(f_mul(radius, sin_angle));
    pos[LEFT][MID][i].y = pos[LEFT][NEAR][i].y +
                          F2I(f_mul(radius, cos_angle));

    double factor = (double)((FLAP_STEPS) - (abs(i - flap_mid) * MID_SPEED)) / 10;
    angle += (d_inc * factor) * MID_SPREAD;
  }

  // Calculate the TIP wing positions
  st_angle = TIP_ST_ANGLE;
  d_inc = tip_flap_inc;
  for (i = 0, angle = st_angle; i < FLAP_STEPS; i++)
  {
    fixed cos_angle, sin_angle;
    viewer->cos_sin(angle, &cos_angle, &sin_angle);

    fixed radius = I2F(wing_span_3 - wing_span_2);
    pos[RIGHT][TIP][i].x = pos[RIGHT][MID][i].x +
                           F2I(f_mul(radius, sin_angle));
    pos[RIGHT][TIP][i].y = pos[RIGHT][MID][i].y +
                           F2I(f_mul(radius, cos_angle));
    pos[LEFT][TIP][i].x = pos[LEFT][MID][i].x -
                          F2I(f_mul(radius, sin_angle));
    pos[LEFT][TIP][i].y = pos[LEFT][MID][i].y +
                          F2I(f_mul(radius, cos_angle));

    double factor = (double)((FLAP_STEPS) - (abs(i - flap_mid) * TIP_SPEED)) / 10;
    angle += (d_inc * factor) * TIP_SPREAD;
  }

  // Set the wing tips
  set_wings();

  // Do a mass color assignment for now
  for (int j = 0; j < TOT_FACES; j++)
  {
    poly[j].top_shade = EARTH_LT_GRAY;
    poly[j].under_shade = 6;
  }
}

//-------------------------------------------------------------------------
// Set the BIRD's wings to the current value of wing_pos
//-------------------------------------------------------------------------
void Bird::set_wings()
{
  // Set the Right wing position
  vert[9].world.x  = vert[10].world.x = pos[RIGHT][NEAR][flap_cnt].x;
  vert[9].world.y  = vert[10].world.y = pos[RIGHT][NEAR][flap_cnt].y;
  vert[11].world.x = vert[12].world.x = pos[RIGHT][MID][flap_cnt].x;
  vert[11].world.y = vert[12].world.y = pos[RIGHT][MID][flap_cnt].y;
  vert[13].world.x = pos[RIGHT][TIP][flap_cnt].x;
  vert[13].world.y = pos[RIGHT][TIP][flap_cnt].y;

  // Set the Left wing position
  vert[16].world.x = vert[17].world.x = pos[LEFT][NEAR][flap_cnt].x;
  vert[16].world.y = vert[17].world.y = pos[LEFT][NEAR][flap_cnt].y;
  vert[18].world.x = vert[19].world.x = pos[LEFT][MID][flap_cnt].x;
  vert[18].world.y = vert[19].world.y = pos[LEFT][MID][flap_cnt].y;
  vert[20].world.x = pos[LEFT][TIP][flap_cnt].x;
  vert[20].world.y = pos[LEFT][TIP][flap_cnt].y;
}

//-------------------------------------------------------------------------
// Drift the BIRD along the world Z axis by altering all its coordinates
//-------------------------------------------------------------------------
void Bird::drift(int distance)
{
 int i, j, k;

 #ifdef DRIFTER
 // Alter the main body vertices
 for (i = 0; i < TOT_VERTICES; i++)
  vert[i].world.z += distance;

 // Alter the wing position vertices
 for (i = 0; i < RIGHT; i++)
  for (j = 0; j < TIP; j++)
   for (k = 0; k < FLAP_STEPS; k++)
    pos[i][j][k].z += distance;

 // Alter the center point
 center.z += distance;
 #endif
}

//-------------------------------------------------------------------------
// Set the BIRD's wings at a given position
//-------------------------------------------------------------------------
void Bird::flap_wings()
{
  // Check for changes in direction
  if (flap_direction == UP)
  {
    shift_body(UP);
    if (flap_cnt < (FLAP_STEPS - 1))
      flap_cnt++;
    else
    {
      flap_direction = DOWN;
      flap_cnt = FLAP_STEPS - 2;
    }
  }
  else if (flap_direction == DOWN)
  {
    shift_body(DOWN);
    if (flap_cnt > 0)
      flap_cnt--;
    else
    {
      flap_direction = UP;
      flap_cnt = 1;
    }
  }

  // Process an existing backward GLIDE condition
  else if (flap_direction == WAIT)
  {
    if (speed_chg < BACK_DRIFT)
      speed_chg += SPEED_INC;
    drift((long)speed_chg);
    if (--wait_cnt == 0)
    {
      flap_direction = save_direction;
      flap_cnt = save_cnt;
      speed_chg = -1.0;
      half_distance = center.z / 2;
    }
  }

  // Check for initiation of a backward GLIDE
  if (flap_direction != WAIT)
  {
    if (flap_direction == UP && flap_cnt == flap_mid + 1 && !random(GLIDE_ODDS))
    {
      #ifdef FLAP_WAIT
      save_direction = flap_direction;
      save_cnt = flap_cnt;
      wait_cnt = GLIDE_WAIT;
      flap_direction = WAIT;
      speed_chg = 1.0;
      #endif
    }
    else
    {
      set_wings();
      if (center.z > 0)
      {
        if (center.z > half_distance)
          speed_chg -= SPEED_INC;
        else if (speed_chg < -1.0)
          speed_chg += SPEED_INC;
        drift((long)speed_chg);
      }
    }
  }
}

//-------------------------------------------------------------------------
// Adjust the bird body up and down as the wings flap
//-------------------------------------------------------------------------
void Bird::shift_body(int direction)
{
  #ifdef SHIFTER
  long inc = direction == UP ? shift_inc : - shift_inc;

  vert[0].world.y += inc;
  vert[1].world.y += inc;
  vert[2].world.y += inc;
  vert[3].world.y += inc;
  vert[15].world.y += inc;
  vert[14].world.y += inc;
  vert[4].world.y += inc;
  vert[5].world.y += inc;
  vert[6].world.y += inc;
  vert[7].world.y += inc;
  vert[8].world.y += inc;
  #endif
}

#ifdef KEY_COMMANDS
//-------------------------------------------------------------------------
// Adjust the entire bird by the given direction, and the given increment
//-------------------------------------------------------------------------
void Bird::elevation(int direction, long s_inc)
{
  int i, j, k;
  long inc = direction == UP ? s_inc : - s_inc;

  for (i = 0; i < TOT_VERTICES; i++)
    vert[i].world.y += inc;
  for (i = 0; i < 2; i++)
    for (j = 0; j < 3; j++)
      for (k = 0; k < FLAP_STEPS; k++)
        pos[i][j][k].y += inc;
  center.y += inc;
}
#endif

//-------------------------------------------------------------------------
// Calculate the screen coordinates of a triangular polygon
//-------------------------------------------------------------------------
void Bird::translate_vertices()
{
  int page = graphics->active_page;

  // Do a z_val translate on the center vertex
  viewer->translate(&vert[0].world, &vert[0].pix[page], &curr_z);

  for (int i = 1; i < TOT_VERTICES; i++)
    viewer->translate(&vert[i].world, &vert[i].pix[page]);
}

//-------------------------------------------------------------------------
// Calculate the bounding box to facilitate a fast screen erase
//-------------------------------------------------------------------------
void Bird::bounding_box()
{
  int page = graphics->active_page;

  // Prepare the bounding box structure
  memcpy(&b_box, &st_box, sizeof(b_box));

  // Scan all vertices
  for (int i = 0; i < TOT_VERTICES; i++)
  {
    i_screen *pix = &vert[i].pix[page];

    // See if this coordinate can update the bounding box
    if (pix->x <= b_box[0].x)
      b_box[0].x = pix->x;
    if (pix->x >= b_box[1].x)
      b_box[1].x = pix->x;
    if (pix->y <= b_box[0].y)
      b_box[0].y = pix->y;
    if (pix->y >= b_box[1].y)
      b_box[1].y = pix->y;
  }
}

//-------------------------------------------------------------------------
// Draw a triangle on screen
//-------------------------------------------------------------------------
void Bird::draw_faces()
{
  int page = graphics->active_page;
  s_point pts[4];

  // Draw all faces
  for (int index = 0; index < TOT_FACES; index++)
  {
    // Fill the polygon endpoint array
    for (int i = 0; i < 3; i++)
      memcpy(&pts[i], &vert[poly[index].v_index[i]].pix[page], 4);

    // Call the triangle fill function
    graphics->solid_fill(pts, 3, poly[index].top_shade);
  }
}

//-------------------------------------------------------------------------
// Display the bird
//-------------------------------------------------------------------------
void Bird::draw()
{
  // Translate all vertices
  translate_vertices();

  // Sort the faces in order of distance
  // sort_faces();

  // Draw all object polygons
  draw_faces();
}

//-------------------------------------------------------------------------
// Erase the bird
//-------------------------------------------------------------------------
void Bird::erase()
{
  bounding_box();
  bar(b_box[0].x, b_box[0].y, b_box[1].x, b_box[1].y);
}
#endif
